open Sdlevent
open Sdlkey
open Sdlvideo
open Program
open Display

let char_to_int c =
    let res = int_of_char c - int_of_char '0' in
    if res < 0 || res > 9 then failwith "Bad move format: a number was expected."
    else res;;

let to_array s =
    let size = ref 1 in
        for i = 0 to (String.length s) - 1 do
            if s.[i] = '/' then incr size
        done;
    let res = Array.make !size (0,0) in
        let buff1 = ref 0 and buff2 = ref 0
        and mult1 = ref 1 and mult2 = ref 1
        and buff_num = ref 1 and move_num = ref 0 in
        for i = 0 to (String.length s) - 1 do
            if s.[i] = '/' then begin
                res.(!move_num) <- (!mult1 * !buff1,!mult2 * !buff2);
                buff1 := 0; buff2 := 0; buff_num := 1;
                mult1 := 1; mult2 := 1;
                incr move_num;
            end
            else if s.[i] = ',' then begin
                res.(!move_num) <- (!buff1,!buff2);
                buff_num := 2;
            end
            else if s.[i] = '(' || s.[i] = ')' then ()
            else if s.[i] = '-' then
                if !buff_num = 1 then mult1 := (-1) else mult2 := (-1)
            else if !buff_num = 1 then buff1 := 10 * !buff1 + (char_to_int s.[i])
            else buff2 := 10 * !buff2 + (char_to_int s.[i])
        done;
        res.(!move_num) <- (!mult1 * !buff1,!mult2 * !buff2);
    res;;

let to_position s =
    let buff1 = ref 0 and buff2 = ref 0
    and mult1 = ref 1 and mult2 = ref 1
    and buff_num = ref 1 in
    for i = 0 to (String.length s) - 1 do
        if s.[i] = ',' then buff_num := 2
        else if s.[i] = '(' || s.[i] = ')' then ()
        else if s.[i] = '-' then
            if !buff_num = 1 then mult1 := (-1) else mult2 := (-1)
        else if !buff_num = 1 then buff1 := 10 * !buff1 + (char_to_int s.[i])
        else buff2 := 10 * !buff2 + (char_to_int s.[i])
        done;
    (!mult1 * !buff1,!mult2 * !buff2);;

let print_moves tab =
    for i = 0 to (Array.length tab) - 1 do
        let (x,y) = tab.(i) in
            Printf.printf "(%d,%d)" x y;
            if i < (Array.length tab) - 1 then print_string "/"
    done;;

let solve adam eve init_pos nb_iter mode =
    print_string "Adam's moves: "; print_moves adam; print_newline ();
    print_string "Eve's moves: "; print_moves eve; print_newline ();
    print_endline "
How to use the program.
The winning set is computed as an attractor,
and after each iteration the new winning positions are displayed, then:
- Esc to quit;
- up and down arrows to zoom in or out (3 levels);
- a to show or hide the axes;
- g to show or hide the grid (at your own risk);
- n to go on;
- left and right arrows to navigate between the computed iterations.
In case of problem, please report to reichert@lsv.ens-cachan.fr.";
(*    let l = compute adam eve nb_iter in
        display_solutions l red*)
    let mode_bool = (if mode = "N" || mode = "n" then 1 else if mode = "Z" || mode = "z" then 2 else 0) in 
    try compute_and_display adam eve init_pos nb_iter mode_bool
    with Failure "stop" -> print_endline "Execution aborted. Bye!"
    | Failure "Found" -> print_endline "The initial position is winning. I stop the execution.";;

let run () =
    let prog_call = Sys.argv in
        if Array.length prog_call = 3 then begin
            let file = open_in prog_call.(1) in
            let adam = input_line file in
            let eve = input_line file in
            let initial_position = input_line file in
            let nb_iter = ref (-1) in
            begin try
            nb_iter := int_of_string (input_line file)
            with Failure "int_of_string" -> ()
            | End_of_file -> ()
            | _ -> failwith "Input error in file while scanning for number of iterations." end;
            close_in file;
            solve (to_array adam) (to_array eve) (to_position initial_position) !nb_iter prog_call.(2)
        end
        else if Array.length prog_call = 4 then begin
            let file = open_in prog_call.(1) in
            let adam = input_line file in
            let eve = input_line file in
            let nb_iter = ref (-1) in
            begin try
            nb_iter := int_of_string (input_line file)
            with Failure "int_of_string" -> ()
            | End_of_file -> ()
            | _ -> failwith "Input error in file while scanning for number of iterations." end;
            close_in file;
            solve (to_array adam) (to_array eve) (to_position prog_call.(2)) !nb_iter prog_call.(3)
        end
        else if Array.length prog_call = 5 then begin
            let adam = prog_call.(1) and eve = prog_call.(2) in
            solve (to_array adam) (to_array eve) (to_position prog_call.(3)) (-1) prog_call.(4)
        end
        else if Array.length prog_call = 6 then begin
            let adam = prog_call.(1) and eve = prog_call.(2) in
            solve (to_array adam) (to_array eve) (to_position prog_call.(3)) (int_of_string prog_call.(4)) prog_call.(4)
        end
        else failwith "Bad program call.
Syntax: solve_RG <Adam's moves> <Eve's moves> <initial position> [number of iterations] <mode>
or solve_RG <path> [initial position] <mode>
where moves are separated by slashes and written in the first two lines of the file,
the mode can be \"N\", \"Z\" or \"0\" (default), which gives the objective,
and there is an optional number of iterations (default: 10).
If the initial position is not specified in the program call,
then it should be the third line of the input file.
Example: solve_RG \"(1,2)/(3,0)\" \"(2,1)/(0,3)\" \"(4,2)\" \"N\".";;

let main () =
    Sdl.init [`VIDEO];
    at_exit Sdl.quit;
    Sdlttf.init ();
    at_exit Sdlttf.quit;
    Sdlkey.enable_unicode true;
    Sdlkey.enable_key_repeat ();
    run ();;

let _ = try main ()
    with | Failure e -> failwith e;
    | e -> raise e
(*    | _ -> failwith "Unexpected failure. Please report bug to reichert@lsv.ens-cachan.fr for correction."*)
