(* open Graphics *)
open Unix
open Sdlevent
open Sdlkey
open Sdlmouse
open Sdlvideo
open Program


let in_array tab elt =
    try
        for i = 0 to (Array.length tab) - 1 do
            if tab.(i) = elt then failwith "Found"
        done; false
    with Failure "Found" -> true;;

let switch tab elt =
    if in_array tab elt then begin
        let res = Array.make ((Array.length tab) - 1) (0,0) and index = ref 0 in
            for i = 0 to (Array.length tab) - 1 do
                if (not (tab.(i) = elt)) then begin
                    res.(!index) <- tab.(i);
                    incr index
                end
            done; res
        end
    else begin
        let res = Array.make ((Array.length tab) + 1) (0,0) in
            for i = 0 to (Array.length tab) - 1 do
                res.(i+1) <- tab.(i);
            done;
            res.(0) <- elt;
            res
    end;;

let print_solutions l =
  let rec print_solutions_rec = function
    | [] -> ()
    | ((x,y)::l) -> Printf.printf "Solution: (%d,%d).\n" x y; print_solutions_rec l
  in print_solutions_rec (List.rev l);;

let grey = (64,64,64);;
let light_grey = (192,192,192);;
let my_blue = (0,0,255);;

let draw_grid zoom screen = match zoom with
  | 1 ->
    for i = 1 to 89 do
      for j = 0 to 750 do
        put_pixel screen (10*i) j (map_RGB screen light_grey)
      done
    done;
    for j = 1 to 74 do
      for i = 0 to 900 do
        put_pixel screen i (10*j+5) (map_RGB screen light_grey)
      done
    done;
  | 2 ->
    for i = 1 to 179 do
      for j = 0 to 750 do
        put_pixel screen (5*i) j (map_RGB screen light_grey)
      done
    done;
    for j = 1 to 149 do
      for i = 0 to 900 do
        put_pixel screen i (5*j) (map_RGB screen light_grey)
      done
    done;
  | 3 -> ()
  | _ -> failwith "Anormal match case in draw_grid";;

let draw_axes zoom screen =
  for i = 0 to 900 do put_pixel screen i 375 (map_RGB screen grey) done;
  for i = 0 to 750 do put_pixel screen 450 i (map_RGB screen grey) done;
  if zoom = 1 then begin
    for i = 0 to 90 do put_pixel screen (10*i) 376 (map_RGB screen grey) done;
    for i = 0 to 74 do put_pixel screen 451 (10*i+5) (map_RGB screen grey) done;
    for i = 0 to 90 do put_pixel screen (10*i) 374 (map_RGB screen grey) done;
    for i = 0 to 74 do put_pixel screen 449 (10*i+5) (map_RGB screen grey) done;
    for i = 0 to 90 do put_pixel screen (10*i) 377 (map_RGB screen grey) done;
    for i = 0 to 74 do put_pixel screen 452 (10*i+5) (map_RGB screen grey) done end;
  flip screen;;

let prepare_display_solutions show_axes show_grid zoom screen =
  fill_rect screen (map_RGB screen white);
  if show_axes then draw_axes zoom screen;
  if show_grid then draw_grid zoom screen;
  flip screen;;

(*    let font = Sdlttf.open_font "font.ttf" 8 in
        let text = Sdlttf.render_text_blended font "Test" ~fg:black in
            blit_surface ~src:text ~dst:screen ();
            flip screen; *)

let put_many_pixels screen x y colour =
  put_pixel screen x y colour;
  put_pixel screen (x+1) y colour;
  put_pixel screen (x-1) y colour;
  put_pixel screen (x+2) y colour;
  put_pixel screen (x-2) y colour;
  put_pixel screen x (y+1) colour;
  put_pixel screen (x+1) (y+1) colour;
  put_pixel screen (x-1) (y+1) colour;
  put_pixel screen (x+2) (y+1) colour;
  put_pixel screen (x-2) (y+1) colour;
  put_pixel screen x (y-1) colour;
  put_pixel screen (x+1) (y-1) colour;
  put_pixel screen (x-1) (y-1) colour;
  put_pixel screen (x+2) (y-1) colour;
  put_pixel screen (x-2) (y-1) colour;
  put_pixel screen x (y+2) colour;
  put_pixel screen (x+1) (y+2) colour;
  put_pixel screen (x-1) (y+2) colour;
  put_pixel screen (x+2) (y+2) colour;
  put_pixel screen (x-2) (y+2) colour;
  put_pixel screen x (y-2) colour;
  put_pixel screen (x+1) (y-2) colour;
  put_pixel screen (x-1) (y-2) colour;
  put_pixel screen (x+2) (y-2) colour;
  put_pixel screen (x-2) (y-2) colour;;

let put_pixels screen x y colour =
  put_pixel screen x y colour;
  put_pixel screen (x+1) y colour;
  put_pixel screen (x-1) y colour;
  put_pixel screen x (y+1) colour;
  put_pixel screen (x+1) (y+1) colour;
  put_pixel screen (x-1) (y+1) colour;
  put_pixel screen x (y-1) colour;
  put_pixel screen (x+1) (y-1) colour;
  put_pixel screen (x-1) (y-1) colour;;

let display_solutions l colour =
(*    print_solutions l; *)
  let screen = set_video_mode 900 750 [`DOUBLEBUF] in
  prepare_display_solutions true false 1 screen;
    let rec display_by_waves = function
      | [] -> ()
      | ((0,0)::tl) -> flip screen;
        print_endline "Iteration completed. Press any key to continue.";
        begin let rec continue () = match wait_event () with
          | KEYDOWN {ke_state=PRESSED} -> display_by_waves tl
          | _ -> continue ()
        in continue () end
      | ((x,y)::tl) -> if (x < 45 && x > (-45) && y < 38 && y > (-38))
        then begin
          put_pixels screen (450+10*x) (375-10*y) (map_RGB screen colour);
        end
        (* else Printf.printf "Winning position out of screen : (%d,%d)\n" x y *);
        display_by_waves tl
    in display_by_waves (List.tl (List.rev l));;

let make_dark total current (x,y,z) =
  (x * (total + 3 * current) / (4 * total),
  y * (total + 3 * current) / (4 * total),
  z * (total + 3 * current) / (4 * total));;

let display_solutions_at_once screen l zoom num_iter colour =
let rec display_solutions_at_once_rec iter colour = function
  | [] -> flip screen
  | ((0,0)::tl) -> display_solutions_at_once_rec (iter-1) colour tl
  | ((x,y)::tl) when zoom = 1 ->
    if (x < 45 && x > (-45) && y < 38 && y > (-38))
    then begin
      put_many_pixels screen (450+10*x) (375-10*y) (map_RGB screen (make_dark num_iter iter colour));
    end;
    display_solutions_at_once_rec iter colour tl
  | ((x,y)::tl) when zoom = 2 ->
    if (x < 90 && x > (-90) && y < 75 && y > (-75))
    then begin
      put_pixels screen (450+5*x) (375-5*y) (map_RGB screen (make_dark num_iter iter colour));
    end;
    display_solutions_at_once_rec iter colour tl
  | ((x,y)::tl) when zoom = 3 ->
    if (x < 450 && x > (-450) && y < 375 && y > (-375))
    then begin
      put_pixel screen (450+x) (375-y) (map_RGB screen (make_dark num_iter iter colour));
    end;
    display_solutions_at_once_rec iter colour tl
  | _ -> failwith "Anormal match case in display_solutions_at_once"
  in display_solutions_at_once_rec num_iter colour l;;

let dsao = display_solutions_at_once;;

let new_positions l =
  let rec until_sep = function
    | [] -> failwith "Unexpected error in the elements of the list."
    | (0,0)::_ -> []
    | hd::tl -> hd::(until_sep tl)
  in until_sep (List.tl l);;

let rec remove_iterations current max = function
  | ((0,0)::tl) when current = max -> (0,0)::tl
  | ((0,0)::tl) -> remove_iterations (current+1) max tl
  | (_::tl) -> remove_iterations current max tl
  | [] -> failwith "Error with the navigation between iterations.";;

let show_moves adam eve =
  Graphics.open_graph " 704x252";
  Graphics.set_window_title "Adam's (left) and Eve's (right) moves";
  Graphics.set_color (Graphics.rgb 64 64 64);
  Graphics.moveto 0 126;
  Graphics.lineto 252 126;
  Graphics.moveto 126 0;
  Graphics.lineto 126 252;
  Graphics.moveto 452 126;
  Graphics.lineto 704 126;
  Graphics.moveto 578 0;
  Graphics.lineto 578 252;
  for i = 0 to 50 do Graphics.plot (5*i+1) 127 done;
  for i = 0 to 50 do Graphics.plot 127 (5*i+1) done;
  for i = 0 to 50 do Graphics.plot (5*i+453) 127  done;
  for i = 0 to 50 do Graphics.plot 579 (5*i+1) done;
  Graphics.set_color Graphics.blue;
  for i = 0 to (Array.length adam) - 1 do
    let (x,y) = adam.(i) in
      if x >= -25 && x <= 25 && y >= -25 && y <= 25 then begin
      Graphics.plot (126+5*x) (126+5*y);
      Graphics.plot (126+5*x) (125+5*y);
      Graphics.plot (126+5*x) (127+5*y);
      Graphics.plot (125+5*x) (126+5*y);
      Graphics.plot (125+5*x) (125+5*y);
      Graphics.plot (125+5*x) (127+5*y);
      Graphics.plot (127+5*x) (126+5*y);
      Graphics.plot (127+5*x) (125+5*y);
      Graphics.plot (127+5*x) (127+5*y)
      end else Printf.printf "Adam's move (%d,%d) could not be represented.\n" x y
  done;
  Graphics.set_color Graphics.red;
  for i = 0 to (Array.length eve) - 1 do
    let (x,y) = eve.(i) in
      if x >= -25 && x <= 25 && y >= -25 && y <= 25 then begin
      Graphics.plot (578+5*x) (126+5*y);
      Graphics.plot (578+5*x) (125+5*y);
      Graphics.plot (578+5*x) (127+5*y);
      Graphics.plot (577+5*x) (126+5*y);
      Graphics.plot (577+5*x) (125+5*y);
      Graphics.plot (577+5*x) (127+5*y);
      Graphics.plot (579+5*x) (126+5*y);
      Graphics.plot (579+5*x) (125+5*y);
      Graphics.plot (579+5*x) (127+5*y)
      end else Printf.printf "Eve's move (%d,%d) could not be represented.\n" x y
  done;;

let mousex_to_coord zoom mousex =
  if zoom = 1 then
    if mousex < 450 then (mousex - 455) / 10 else (mousex - 445) / 10
  else if zoom = 2 then
    if mousex < 450 then (mousex - 453) / 5 else (mousex - 448) / 5
  else mousex - 450;;

let mousey_to_coord zoom mousey =
  if zoom = 1 then
    if mousey < 375 then (380 - mousey) / 10 else (370 - mousey) / 10
  else if zoom = 2 then
    if mousey < 375 then (377 - mousey) / 5 else (372 - mousey) / 5
  else 375 - mousey;;

let find_iteration x y positions =
  try let rec find_iteration_rec x y = function
  | (xx,yy)::_ when x = xx && y = yy -> 0
  | (0,0)::tl -> 1 + find_iteration_rec x y tl
  | _::tl -> find_iteration_rec x y tl
  | [] -> failwith "Not found"
  in find_iteration_rec x y positions
  with Failure "Not found" -> -1;;

let rec find_sum (x,y) = function
  | [] -> ([],false)
  | (x1,y1)::tl when x=x1 && y=y1 -> find_sum (x,y) tl
  | (0,0)::tl -> find_sum (x,y) tl
  | (x1,y1)::tl -> if List.mem (x-x1,y-y1) ((x1,y1)::tl) then ([(x1,y1);(x-x1,y-y1)],true) else find_sum (x,y) tl;;

let rec compute_and_display adam eve nb_iter =
  let screen = set_video_mode 900 750 [`DOUBLEBUF] in Sdlwm.set_caption ~title:"Eve's winning set" ~icon:"";
  show_moves adam eve;
  prepare_display_solutions true false 1 screen;
  let rec pre_n set_x zoom_level show_axes show_grid iter =
    match iter with
    | 0 -> print_endline "End of iterations. Press any key to leave.";
      begin let rec continue () = match wait_event () with
        | KEYDOWN {ke_state=PRESSED} -> ()
        | _ -> continue ()
      in continue () end
    | n -> Printf.printf "Iteration %d..." (nb_iter-n+1); flush_all ();
      let next_set = attr adam eve set_x in
      let new_pos = new_positions next_set and currently_shown = ref nb_iter and move_adam = ref 0 in
        display_solutions_at_once screen next_set zoom_level (nb_iter-n+1) my_blue;
        let new_zoom = ref zoom_level and new_axes = ref show_axes and new_grid = ref show_grid in
          let rec display_wave = function
            | [] -> flip screen; let who_switches = ref 0 in
              begin let rec continue () = match wait_event () with
                | KEYDOWN {keysym=KEY_ESCAPE} -> failwith "stop"
                | KEYDOWN {keysym=KEY_UP} ->
                  new_zoom := min 3 (!new_zoom + 1);
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keysym=KEY_DOWN} ->
                  new_zoom := max 1 (!new_zoom - 1);
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keysym=KEY_LEFT} ->
                  currently_shown := max (!currently_shown-1) n;
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keysym=KEY_RIGHT} ->
                  currently_shown := min (!currently_shown+1) nb_iter;
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keycode='a'} -> new_axes := not !new_axes;
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keycode='g'} -> new_grid := not !new_grid;
                  let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                    prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                    dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                    display_wave (new_positions set_to_show)
                | KEYDOWN {keycode='w'} -> who_switches := 1 - !who_switches; continue ();
                | KEYDOWN {keycode='z'} -> if !who_switches = 1 then
                    let adam2 = switch adam (-1,1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (-1,1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='e'} -> if !who_switches = 1 then
                    let adam2 = switch adam (0,1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (0,1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='r'} -> if !who_switches = 1 then
                    let adam2 = switch adam (1,1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (1,1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='s'} -> if !who_switches = 1 then
                    let adam2 = switch adam (-1,0) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (-1,0) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='d'} -> if !who_switches = 1 then
                    let adam2 = switch adam (0,0) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (0,0) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='f'} -> if !who_switches = 1 then
                    let adam2 = switch adam (1,0) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (1,0) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='x'} -> if !who_switches = 1 then
                    let adam2 = switch adam (-1,-1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (-1,-1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='c'} -> if !who_switches = 1 then
                    let adam2 = switch adam (0,-1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (0,-1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='v'} -> if !who_switches = 1 then
                    let adam2 = switch adam (1,-1) in compute_and_display adam2 eve nb_iter
                  else let eve2 = switch eve (1,-1) in
                    compute_and_display adam eve2 nb_iter
                | KEYDOWN {keycode='n'} -> ()
                | MOUSEBUTTONDOWN {mbe_button=BUTTON_MIDDLE; mbe_x=_ as mousex; mbe_y=_ as mousey}
                when cursor_visible () ->
                  let x = mousex_to_coord !new_zoom mousex and y = mousey_to_coord !new_zoom mousey in
                    if x = 0 && y = 0 then continue () else begin
                      let num_iter = find_iteration x y next_set in
                        currently_shown := (if num_iter = -1 then nb_iter else nb_iter-num_iter+1);
                        let set_to_show = remove_iterations !currently_shown nb_iter next_set in
                          prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                          dsao screen set_to_show !new_zoom (nb_iter-n+1) my_blue;
                          display_wave (new_positions set_to_show)
                      end
                | MOUSEBUTTONDOWN {mbe_button=BUTTON_RIGHT; mbe_x=_ as mousex; mbe_y=_ as mousey}
                when cursor_visible () ->
                  let x = mousex_to_coord !new_zoom mousex and y = mousey_to_coord !new_zoom mousey in
                    if (x = 0 && y = 0) || (not (List.mem (x,y) next_set)) then continue () else
                      begin let num_iter = find_iteration x y next_set in
                        currently_shown := nb_iter-num_iter+1;
                        let set_to_show = remove_iterations (!currently_shown-1) nb_iter next_set in
                          prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                          let (pair,answer) = find_sum (x,y) set_to_show in
                            if answer then begin
                              let (x1,y1) = List.hd pair and (x2,y2) = List.hd (List.tl pair) in
                                if !new_zoom = 1 then begin
                                  put_many_pixels screen (450+10*x1) (375-10*y1) (map_RGB screen blue);
                                  put_many_pixels screen (450+10*x2) (375-10*y2) (map_RGB screen blue) end
                                else if !new_zoom = 2 then begin
                                  put_pixels screen (450+5*x1) (375-5*y1) (map_RGB screen blue);
                                  put_pixels screen (450+5*x2) (375-5*y2) (map_RGB screen blue) end
                                else begin
                                  put_pixel screen (450+x1) (375-y1) (map_RGB screen blue);
                                  put_pixel screen (450+x2) (375-y2) (map_RGB screen blue) end;
                                display_wave [(x,y)]
                            end
                            else begin
                              let s_t_s = remove_iterations !currently_shown nb_iter next_set in
                                dsao screen s_t_s !new_zoom (nb_iter-n+1) my_blue;
                                display_wave (new_positions s_t_s)
                            end
                      end
                | MOUSEBUTTONDOWN {mbe_button=BUTTON_WHEELUP; mbe_x=_ as mousex; mbe_y=_ as mousey}
                when cursor_visible () ->
                  incr move_adam;
                  if !move_adam = Array.length adam then move_adam := 0;
                  let x = mousex_to_coord !new_zoom mousex and y = mousey_to_coord !new_zoom mousey in
                    if (x = 0 && y = 0) || (not (List.mem (x,y) next_set)) then continue () else
                      begin let num_iter = find_iteration x y next_set in
                        currently_shown := nb_iter-num_iter+1;
                        let set_to_show = remove_iterations (!currently_shown-1) nb_iter next_set in
                          prepare_display_solutions !new_axes !new_grid !new_zoom screen;
                          let (adamx,adamy) = adam.(!move_adam) in
                            if !new_zoom = 1 then put_many_pixels screen (450+10*(x+adamx)) (375-10*(y+adamy)) (map_RGB screen black)
                            else if !new_zoom = 2 then put_pixels screen (450+5*(x+adamx)) (375-5*(y+adamy)) (map_RGB screen black)
                            else put_pixel screen (450+x+adamx) (375-y-adamy) (map_RGB screen black);
                            let index = ref 0 and found = ref false in
                              try while not !found do
                                let (evex,evey) = eve.(!index) in
                                  if List.mem (x+adamx+evex,y+adamy+evey) set_to_show then begin
                                    found := true;
                                    if !new_zoom = 1 then put_many_pixels screen (450+10*(x+adamx+evex)) (375-10*(y+adamy+evey)) (map_RGB screen blue)
                                    else if !new_zoom = 2 then put_pixels screen (450+5*(x+adamx+evex)) (375-5*(y+adamy+evey)) (map_RGB screen blue)
                                    else put_pixel screen (450+x+adamx+evex) (375-y-adamy-evey) (map_RGB screen blue);
                                  end
                                  else incr index
                              done; display_wave [(x,y)]
                              with Invalid_argument(e) -> failwith e 
                      end
                | _ -> continue ()
              in continue () end
            | ((x,y)::tl) when !new_zoom = 1 -> if (x < 45 && x > (-45) && y < 38 && y > (-38))
              then put_many_pixels screen (450+10*x) (375-10*y) (map_RGB screen red);
              (* else Printf.printf "Winning position out of screen : (%d,%d)\n" x y *)
              display_wave tl
            | ((x,y)::tl) when !new_zoom = 2 -> if (x < 90 && x > (-90) && y < 75 && y > (-75))
              then put_pixels screen (450+5*x) (375-5*y) (map_RGB screen red);
              (* else Printf.printf "Winning position out of screen : (%d,%d)\n" x y *)
              display_wave tl
            | ((x,y)::tl) when !new_zoom = 3 -> if (x < 450 && x > (-450) && y < 375 && y > (-375))
              then put_pixel screen (450+x) (375-y) (map_RGB screen red);
              (* else Printf.printf "Winning position out of screen : (%d,%d)\n" x y *)
              display_wave tl
            | _ -> failwith "Anormal match case in compute_and_display"
          in print_endline " completed.";
          display_wave new_pos;
          pre_n next_set !new_zoom !new_axes !new_grid (n-1)
    in pre_n [(0,0)] 1 true false nb_iter;;

